package org.apache.bcel.classfile;

import java.io.DataOutputStream;
import java.util.Arrays;
import org.apache.commons.lang3.ArrayUtils;
import java.io.IOException;
import java.io.DataInput;

public final class Code extends Attribute
{
    private int maxStack;
    private int maxLocals;
    private byte[] code;
    private CodeException[] exceptionTable;
    private Attribute[] attributes;
    
    public Code(final Code c) {
        this(c.getNameIndex(), c.getLength(), c.getMaxStack(), c.getMaxLocals(), c.getCode(), c.getExceptionTable(), c.getAttributes(), c.getConstantPool());
    }
    
    Code(final int nameIndex, final int length, final DataInput file, final ConstantPool constantPool) throws IOException {
        this(nameIndex, length, file.readUnsignedShort(), file.readUnsignedShort(), null, null, null, constantPool);
        final int codeLength = file.readInt();
        file.readFully(this.code = new byte[codeLength]);
        final int exceptionTableLength = file.readUnsignedShort();
        this.exceptionTable = new CodeException[exceptionTableLength];
        for (int i = 0; i < exceptionTableLength; ++i) {
            this.exceptionTable[i] = new CodeException(file);
        }
        final int attributesCount = file.readUnsignedShort();
        this.attributes = new Attribute[attributesCount];
        for (int j = 0; j < attributesCount; ++j) {
            this.attributes[j] = Attribute.readAttribute(file, constantPool);
        }
        super.setLength(length);
    }
    
    public Code(final int nameIndex, final int length, final int maxStack, final int maxLocals, final byte[] code, final CodeException[] exceptionTable, final Attribute[] attributes, final ConstantPool constantPool) {
        super((byte)2, nameIndex, length, constantPool);
        this.maxStack = maxStack;
        this.maxLocals = maxLocals;
        this.code = ((code != null) ? code : ArrayUtils.EMPTY_BYTE_ARRAY);
        this.exceptionTable = ((exceptionTable != null) ? exceptionTable : CodeException.EMPTY_CODE_EXCEPTION_ARRAY);
        this.attributes = ((attributes != null) ? attributes : Code.EMPTY_ARRAY);
        super.setLength(this.calculateLength());
    }
    
    @Override
    public void accept(final Visitor v) {
        v.visitCode(this);
    }
    
    private int calculateLength() {
        int len = 0;
        if (this.attributes != null) {
            for (final Attribute attribute : this.attributes) {
                len += attribute.getLength() + 6;
            }
        }
        return len + this.getInternalLength();
    }
    
    @Override
    public Attribute copy(final ConstantPool constantPool) {
        final Code c = (Code)this.clone();
        if (this.code != null) {
            c.code = this.code.clone();
        }
        c.setConstantPool(constantPool);
        Arrays.setAll(c.exceptionTable = new CodeException[this.exceptionTable.length], i -> this.exceptionTable[i].copy());
        Arrays.setAll(c.attributes = new Attribute[this.attributes.length], i -> this.attributes[i].copy(constantPool));
        return c;
    }
    
    @Override
    public void dump(final DataOutputStream file) throws IOException {
        super.dump(file);
        file.writeShort(this.maxStack);
        file.writeShort(this.maxLocals);
        file.writeInt(this.code.length);
        file.write(this.code, 0, this.code.length);
        file.writeShort(this.exceptionTable.length);
        for (final CodeException exception : this.exceptionTable) {
            exception.dump(file);
        }
        file.writeShort(this.attributes.length);
        for (final Attribute attribute : this.attributes) {
            attribute.dump(file);
        }
    }
    
    public Attribute[] getAttributes() {
        return this.attributes;
    }
    
    public byte[] getCode() {
        return this.code;
    }
    
    public CodeException[] getExceptionTable() {
        return this.exceptionTable;
    }
    
    private int getInternalLength() {
        return 8 + this.code.length + 2 + 8 * ((this.exceptionTable == null) ? 0 : this.exceptionTable.length) + 2;
    }
    
    public LineNumberTable getLineNumberTable() {
        for (final Attribute attribute : this.attributes) {
            if (attribute instanceof LineNumberTable) {
                return (LineNumberTable)attribute;
            }
        }
        return null;
    }
    
    public LocalVariableTable getLocalVariableTable() {
        for (final Attribute attribute : this.attributes) {
            if (attribute instanceof LocalVariableTable) {
                return (LocalVariableTable)attribute;
            }
        }
        return null;
    }
    
    public int getMaxLocals() {
        return this.maxLocals;
    }
    
    public int getMaxStack() {
        return this.maxStack;
    }
    
    public void setAttributes(final Attribute[] attributes) {
        this.attributes = ((attributes != null) ? attributes : Code.EMPTY_ARRAY);
        super.setLength(this.calculateLength());
    }
    
    public void setCode(final byte[] code) {
        this.code = ((code != null) ? code : ArrayUtils.EMPTY_BYTE_ARRAY);
        super.setLength(this.calculateLength());
    }
    
    public void setExceptionTable(final CodeException[] exceptionTable) {
        this.exceptionTable = ((exceptionTable != null) ? exceptionTable : CodeException.EMPTY_CODE_EXCEPTION_ARRAY);
        super.setLength(this.calculateLength());
    }
    
    public void setMaxLocals(final int maxLocals) {
        this.maxLocals = maxLocals;
    }
    
    public void setMaxStack(final int maxStack) {
        this.maxStack = maxStack;
    }
    
    @Override
    public String toString() {
        return this.toString(true);
    }
    
    public String toString(final boolean verbose) {
        final StringBuilder buf = new StringBuilder(100);
        buf.append("Code(maxStack = ").append(this.maxStack).append(", maxLocals = ").append(this.maxLocals).append(", code_length = ").append(this.code.length).append(")\n").append(Utility.codeToString(this.code, super.getConstantPool(), 0, -1, verbose));
        if (this.exceptionTable.length > 0) {
            buf.append("\nException handler(s) = \n").append("From\tTo\tHandler\tType\n");
            for (final CodeException exception : this.exceptionTable) {
                buf.append(exception.toString(super.getConstantPool(), verbose)).append("\n");
            }
        }
        if (this.attributes.length > 0) {
            buf.append("\nAttribute(s) = ");
            for (final Attribute attribute : this.attributes) {
                buf.append("\n").append(attribute.getName()).append(":");
                buf.append("\n").append(attribute);
            }
        }
        return buf.toString();
    }
}
